import { readFile, writeFile, readdir } from "node:fs/promises";
import { spawn } from "node:child_process";
import { resolve } from "node:path";
import { mkdirSync } from "node:fs";
import { rimrafSync } from "rimraf";

rimrafSync("tmp");
mkdirSync("tmp");

rimrafSync("deck");
mkdirSync("deck");

const execAsync = (args) => {
  return new Promise((resolve) => {
    const process = spawn(args.shift(), args, { stdio: "inherit" });
    process.on("exit", resolve);
  });
};

const typst = {
  compile: (input, output) => execAsync(["typst", "compile", input, output]),
};

const header = (title, overrides = "") =>
  `#import "preamble.typ": *\n#set page(width: 360pt, height: auto, margin: (x: 4pt, y: 8pt))\n#set page(header: text(gray, [${title}]))\n${overrides}\n#set-page-properties()\n`;

const tmp = (ext) =>
  resolve("tmp", `${Math.random().toString().substring(2)}.${ext}`);

const RE_NOTE_V1 = /\/\* @card \*\/([^]*?)\/\* --- \*\/([^]*?)\/\* @end \*\//g;
const RE_CLOZE_V1 = /\/\* @cloze \*\/([^]*?)\/\* @end \*\//g;
const RE_NOTE_WITH_TITLE_V1 =
  /\/\* @card ([\w\d\s]+?) \*\/([^]*?)\/\* @end \*\//g;
const V1_EXPCT_CARDS = /\@card|\@cloze/g;

function findLastTitle(s, i) {
  function recMatch(lastIndex, maxLength = 5) {
    if (maxLength <= 0) return [];

    const inp = s.substring(0, lastIndex);
    const RE_TITLE = new RegExp(`^(={1,${maxLength}}) ([^]+?)$`, "gm");

    let lastMatch = null,
      currentMatch = RE_TITLE.exec(inp);
    while (currentMatch !== null) {
      lastMatch = currentMatch;
      currentMatch = RE_TITLE.exec(inp);
    }

    if (lastMatch) {
      return [...recMatch(lastIndex, lastMatch[1].length - 1), lastMatch[2]];
    } else {
      return [];
    }
  }

  return `===== ${recMatch(i).join(" #sym.angle.r ")}`;
}

async function run() {
  await writeFile(
    resolve("tmp", "preamble.typ"),
    await readFile(resolve("src", "preamble.typ"))
  );

  const decks = await readdir("src").then((l) =>
    l.filter((e) => e !== "preamble.typ" && e.endsWith(".typ"))
  );

  for (const fn of decks) {
    const deck_name = fn.replace(/.typ/, "");

    const f = await readFile(resolve("src", fn));
    const s = f.toString();

    let value = RE_NOTE_WITH_TITLE_V1.exec(s);
    let count = 0;
    while (value) {
      const [_, card_title, content] = value;

      const title = `#text(size: 0.75em, [${findLastTitle(
        s,
        RE_NOTE_WITH_TITLE_V1.lastIndex
      )}])`;

      const in1 = tmp("typ"),
        in2 = tmp("typ"),
        out1 = resolve("deck", `${deck_name}_${++count}_front.svg`),
        out2 = resolve("deck", `${deck_name}_${count}_back.svg`);

      const final_title = `#align(center, [${card_title}])`;

      await writeFile(in1, header(title) + final_title);
      await writeFile(in2, header(title) + final_title + "\n\n" + content);

      await typst.compile(in1, out1);
      await typst.compile(in2, out2);

      value = RE_NOTE_WITH_TITLE_V1.exec(s);
    }

    value = RE_NOTE_V1.exec(s);
    while (value) {
      const [_, top, bottom] = value;

      const title = `#text(size: 0.75em, [${findLastTitle(
        s,
        RE_NOTE_V1.lastIndex
      )}])`;

      const in1 = tmp("typ"),
        in2 = tmp("typ"),
        out1 = resolve("deck", `${deck_name}_${++count}_front.svg`),
        out2 = resolve("deck", `${deck_name}_${count}_back.svg`);

      await writeFile(in1, header(title) + top);
      await writeFile(in2, header(title) + top + "\n\n" + bottom);

      await typst.compile(in1, out1);
      await typst.compile(in2, out2);

      value = RE_NOTE_V1.exec(s);
    }

    value = RE_CLOZE_V1.exec(s);
    while (value) {
      const [_, content] = value;

      const title = `#text(size: 0.75em, [${findLastTitle(
        s,
        RE_CLOZE_V1.lastIndex
      )}])`;

      const in1 = tmp("typ"),
        in2 = tmp("typ"),
        out1 = resolve("deck", `${deck_name}_${++count}_front.svg`),
        out2 = resolve("deck", `${deck_name}_${count}_back.svg`);

      await writeFile(
        in1,
        header(
          title,
          "#let cloze(content) = [[...]]\n#let clozeh(content) = []"
        ) + content
      );
      await writeFile(in2, header(title) + content);

      await typst.compile(in1, out1);
      await typst.compile(in2, out2);

      value = RE_CLOZE_V1.exec(s);
    }

    await writeFile(
      resolve("deck", deck_name + ".csv"),

      new Array(count)
        .fill(0)
        .map((_, i) =>
          [
            `<img src='${deck_name}_${
              i + 1
            }_front.svg' width='100%' style='filter: invert(100%)'>`,
            `<img src='${deck_name}_${
              i + 1
            }_back.svg' width='100%' style='filter: invert(100%)'>`,
          ].join("|")
        )
        .join("\n")
    );

    const expected = (s.match(V1_EXPCT_CARDS) ?? 0).length;
    if (expected !== count) {
      console.error(
        `Mismatch for ${deck_name}, expected ${expected}, found ${count}.`
      );
    }
  }
}

run();
