// @ts-check
import { mkdir, readFile, writeFile, stat, copyFile } from "node:fs/promises";
import { spawn } from "node:child_process";
import { dirname, relative, resolve, sep } from "node:path";
import { rimraf } from "rimraf";
import { mkdirp } from "mkdirp";
import hash from "object-hash";
import sharp from "sharp";
import dayjs from "dayjs";
import Pino from "pino";

/**
 * Use pino for logging
 */
const logger = Pino({
  level: "trace",
  transport: {
    target: "pino-pretty",
  },
});

/**
 * Asynchronously execute a process
 * @param {string[]} args Arguments
 * @param {import("node:child_process").SpawnOptions} spawnOptions Spawn options
 * @returns {Promise<never>}
 */
function execAsync(args, spawnOptions = {}) {
  return new Promise((resolve) => {
    const cmd = args.shift() ?? "";
    const process = spawn(cmd, args, { stdio: "inherit", ...spawnOptions });
    process.on("exit", resolve);
  });
}

/**
 * Create directory if it doesn't exist, otherwise clear it.
 * @param {string} path Directory path
 */
async function configureDirectory(path) {
  await stat(path)
    .then(async (f) => {
      if (f.isDirectory()) {
        await rimraf(path + "/*", {
          glob: true,
        });
      } else {
        await rimraf(path);
        await mkdir(path);
      }
    })
    .catch(async (err) => {
      if (err.code === "ENOENT") {
        await mkdir(path);
      }
    });
}

/**
 * Clean up temporary files
 */
async function cleanUp() {
  await rimraf("temp");
}

/**
 * Scan a string to detect the line number of a given match
 * @param {string} str Input string
 * @param {number} lastIndex Last index
 */
function lastIndexToLineNumber(str, lastIndex = 0) {
  let lineNum = 0;
  for (let idx = 0; idx < lastIndex; idx++) {
    if (str.charAt(idx) == "\n") {
      lineNum++;
    }
  }

  return lineNum;
}

/**
 * Regex for matching titles
 */
const RE_TITLE = /^(={1,5}) ([^]+?)$/;

/**
 * Convert Typst source code to titles
 * @param {string} source Input source
 * @returns {[number, number, string][]} Titles
 */
function typstToTitles(source) {
  return source
    .split("\n")
    .map((line, lineNum) => {
      const match = RE_TITLE.exec(line);
      if (match) {
        const [, heading, title] = match;
        return [lineNum, heading.length, title];
      }

      /** @type {any} */
      let fakeType = null;
      return fakeType;
    })
    .filter((x) => x);
}

/**
 * Get a list of current titles from current line number
 * @param {[number, number, string][]} titles List of (line number, heading size, title)
 * @param {number} lineNum Current line number
 * @returns {string[]} Title list
 */
function constructTitleList(titles, lineNum) {
  const candidates = titles
    .filter(([titleLineNum]) => titleLineNum <= lineNum)
    .reverse();
  let currentDepth = 5;

  let titleList = [];
  while (currentDepth > 0) {
    const nextMatch = candidates.find(([, heading]) => heading <= currentDepth);

    if (nextMatch) {
      const [, heading, title] = nextMatch;
      currentDepth = heading - 1;
      titleList.push(title);
    } else {
      currentDepth = 0;
    }
  }

  return titleList.reverse();
}

/**
 * Regex to detect V1 cards
 */
const RE_CARD_V1 =
  /\/\* @card \*\/(?<card_v1_q>[^]*?)\/\* --- \*\/(?<card_v1_a>[^]*?)\/\* @end \*\/|\/\* @cloze \*\/(?<cloze_v1>[^]*?)\/\* @end \*\/|\/\* @card (?<card_v1_title>[\w\d\s]+?) \*\/(?<card_v1_title_body>[^]*?)\/\* @end \*\//g;

/**
 * Extract all V1 cards in orderly data structure
 * @param {string} source Source file
 */
function extractV1Cards(source) {
  /** @type {({ lastIndex: number } & ({ type: 'card_v1', q: string, a: string } | { type: 'cloze_v1', content: string }))[]} */
  const cards = [];

  /** @type {any} */
  let match = RE_CARD_V1.exec(source);
  while (match !== null) {
    /** @type {{ card_v1_q: string, card_v1_a: string, cloze_v1: string, card_v1_title: string, card_v1_title_body: string }} */
    const groups = match.groups;

    if (groups.card_v1_q) {
      // Card V1
      cards.push({
        lastIndex: RE_CARD_V1.lastIndex,
        type: "card_v1",
        q: groups.card_v1_q.trim(),
        a: groups.card_v1_a.trim(),
      });
    } else if (groups.cloze_v1) {
      // Cloze V1
      cards.push({
        lastIndex: RE_CARD_V1.lastIndex,
        type: "cloze_v1",
        content: groups.cloze_v1.trim(),
      });
    } else if (groups.card_v1_title) {
      // Card with Title V1
      cards.push({
        lastIndex: RE_CARD_V1.lastIndex,
        type: "card_v1",
        q: groups.card_v1_title.trim(),
        a: groups.card_v1_title_body.trim(),
      });
    }

    match = RE_CARD_V1.exec(source);
  }

  return cards;
}

/**
 * Regex for matching assets
 */
const RE_ASSET =
  /^#import "(?<typ>.+?).typ".*?$|^#include "(?<inc>.+?).typ"$|image\([^"]*?"(?<png>.+?).(png|jpg)"/gm;

/**
 * Detect required assets and rewrite paths in the given source
 * @param {string} cwd Current directory to resolve from
 * @param {string} source Source file
 * @param {('typ' | 'inc' | 'png')[] | undefined} filter Filter for specific assets
 * @returns List of assets and rewritten source
 */
async function rewriteAssets(cwd, source, filter) {
  /** @type {{ type: 'typ' | 'inc' | 'png', path: string }[]} */
  const assets = [];

  /** @type {any} */
  let match = RE_ASSET.exec(source);
  while (match !== null) {
    /** @type {{ typ: string, png: string }} */
    const groups = match.groups;

    /** @type {any} */
    const type = Object.keys(groups).filter((group) => groups[group])[0];

    if (!filter || filter.includes(type)) {
      if (type === "png") {
        source =
          source.slice(0, RE_ASSET.lastIndex - 'png"'.length) +
          "jpg" +
          source.slice(RE_ASSET.lastIndex - '"'.length);
      }

      const ext = "." + (type === "png" ? "png" : "typ");

      assets.push({
        type,
        path: resolve(cwd, groups[type] + ext),
      });
    }

    match = RE_ASSET.exec(source);
  }

  return { assets, source };
}

/**
 * Generate card sides with hash
 * @param {{ lastIndex?: number } & ({ type: 'card_v1', q: string, a: string } | { type: 'cloze_v1', content: string })} card Card
 */
function generateCardSideSourceWithHash(card) {
  if (card.type === "card_v1") {
    return {
      front: card.q,
      back: `${card.q}\n#linebreak()\n${card.a}`,

      lastIndex: card.lastIndex,
      hash: hash(card.q, { algorithm: "sha256" }).substring(0, 12),
    };
  } else {
    return {
      front: `#let cloze(content) = [[...]]\n#let clozeh(content) = []\n${card.content}`,
      back: card.content,

      lastIndex: card.lastIndex,
      hash: hash(card.content, { algorithm: "sha256" }).substring(0, 12),
    };
  }
}

/**
 * Process a Typst source file
 * @param {string} fp File path of current source file
 * @returns {Promise<{ cards: { titles: string[], front: string, back: string, hash: string, cwd: string }[], assets: { type: string, path: string }[] }>}
 */
async function processTypstSourceFile(fp) {
  fp = resolve(fp);
  const cwd = dirname(fp);

  const f = await readFile(fp);
  const typst = await f.toString();

  const cards = extractV1Cards(typst);
  const titles = typstToTitles(typst);
  const processedCards = await Promise.all(
    cards.map(generateCardSideSourceWithHash).map(async (card) => {
      const { source: front } = await rewriteAssets(cwd, card.front, ["png"]);
      const { assets, source: back } = await rewriteAssets(cwd, card.back, [
        "png",
      ]);

      const lineNumber = lastIndexToLineNumber(typst, card.lastIndex);
      delete card["lastIndex"];

      return {
        titles: constructTitleList(titles, lineNumber),
        front,
        back,
        hash: card.hash,
        assets,
        cwd,
      };
    })
  );

  const subAssets = (await rewriteAssets(cwd, typst, ["typ", "inc"])).assets;
  const subFiles = await Promise.all(
    subAssets
      .filter((asset) => asset.type === "inc")
      .map((asset) => processTypstSourceFile(asset.path))
  );

  const finalCards = [
    ...processedCards,
    ...subFiles.flatMap((file) => file.cards),
  ];

  const fileAssets = [];
  finalCards
    .filter((card) => card["assets"])
    .forEach((card) => {
      fileAssets.push(...card["assets"]);
      delete card["assets"];
    });

  return {
    cards: finalCards,
    assets: [
      ...subAssets.filter((asset) => asset.type === "typ"),
      ...subFiles.flatMap((file) => file.assets),
      ...fileAssets,
    ],
  };
}

/**
 * Copy list of assets from source to target directory
 * @param {string} src Source directory (absolute)
 * @param {string} dest Target directory (absolute)
 * @param {{ type: string, path: string }[]} assets Absolute path names to assets
 */
async function processAssets(src, dest, assets) {
  logger.info("Processing assets...");
  const processed = new Set();

  for (const asset of assets) {
    if (processed.has(asset.path)) continue;

    processed.add(asset.path);
    if (!asset.path.startsWith(src))
      throw `Asset "${asset}" does not start with "${src}", this is unexpected!`;

    const targetFn = resolve(dest, asset.path.slice(src.length + 1));
    await mkdirp(dirname(targetFn));

    const debugFn = asset.path.split(sep).pop();
    if (asset.type === "png") {
      await sharp(asset.path).toFile(
        targetFn.slice(0, targetFn.length - 3) + "jpg"
      );

      logger.debug(`Converted asset ${debugFn}`);
    } else {
      await copyFile(asset.path, targetFn);
      logger.debug(`Copied asset ${debugFn}`);
    }
  }
}

/**
 * Regex to match Typst citation
 */
const RE_CITATION = /@(?!preview)[\w\d_-]+/g;

/**
 * Regex to match Typst footnote
 */
const RE_FOOTNOTE = /#footnote\[[^]+?\]/g;

/**
 * Format content within a Typst source file for a card
 * @param {string[]} titles Titles
 * @param {string} content Content
 * @param {string} preamble Path to preamble file
 * @returns Typst source file
 */
function formatTypstFile(titles, content, preamble) {
  let match = RE_CITATION.exec(content);
  while (match) {
    logger.warn(`Stripping citation: ${match[0]}`);
    match = RE_CITATION.exec(content);
  }

  return `#import "${preamble}": *
#set page(width: 360pt, height: auto, margin: (x: 4pt, y: 8pt))
${
  titles.length &&
  `===== #text(gray, 0.75em, [${titles
    .join(" #sym.angle.r ")
    .replace(RE_FOOTNOTE, "")}])`
}

${content.replace(RE_CITATION, "")}`;
}

/**
 * Generate Typst files for cards
 * @param {string} src Source directory (absolute)
 * @param {string} dest Target directory (absolute)
 * @param {{ titles: string[], front: string, back: string, hash: string, cwd: string }[]} cards
 */
async function generateTypstFiles(src, dest, cards) {
  return Promise.all(
    cards.map(async (card) => {
      const cwd = resolve(dest, card.cwd.slice(src.length + 1));
      const preamble = relative(cwd, resolve(src, "preamble.typ"));

      const front = formatTypstFile(card.titles, card.front, preamble);
      const back = formatTypstFile(card.titles, card.back, preamble);

      if (!card.cwd.startsWith(src))
        throw `Cwd "${card.cwd}" does not start with "${src}", unexpected!`;

      const path = (side) => resolve(cwd, `${card.hash}_${side}.typ`);

      await writeFile(path("front"), front);
      await writeFile(path("back"), back);

      logger.debug(`Wrote card ${card.hash} front and back.`);
    })
  );
}

/**
 * Render a Typst file
 * @param {string} input Input file
 * @param {string} output Output file
 * @param {string} root Root directory
 */
function compileTypstFile(input, output, root) {
  return execAsync(["typst", "compile", input, output], {
    env: {
      ...process.env,
      TYPST_ROOT: root,
    },
  });
}

/**
 * Render given cards to SVGs
 * @param {string} src Source directory (absolute)
 * @param {string} dest Target directory (absolute)
 * @param {string} build Build directory (absolute)
 * @param {{ hash: string, cwd: string }[]} cards
 */
async function renderCards(src, dest, build, cards) {
  return Promise.all(
    cards.flatMap(async (card) => {
      const cwd = resolve(dest, card.cwd.slice(src.length + 1));
      const path = (side) => resolve(cwd, `${card.hash}_${side}.typ`);
      const pathOutput = (side) => resolve(build, `${card.hash}_${side}.svg`);

      return ["front", "back"].map(async (side) => {
        await compileTypstFile(
          path(side),
          pathOutput(side),
          resolve(dest, "../..")
        );

        logger.debug(`Rendered card ${card.hash}.`);
      });
    })
  );
}

/**
 * Write the CSV file for the deck
 * @param {string} build Build directory (absolute)
 * @param {{ hash: string }[]} cards Cards
 * @param {string} deckName Deck name
 */
async function writeDeckCsv(build, cards, deckName) {
  return writeFile(
    resolve(build, deckName),

    cards
      .map((card) =>
        [
          `<img src='${card.hash}_front.svg' width='100%' style='filter: invert(100%)'>`,
          `<img src='${card.hash}_back.svg' width='100%' style='filter: invert(100%)'>`,
        ].join("|")
      )
      .join("\n")
  );
}

/**
 * Process a Typst deck
 * @param {string} src Source directory (absolute)
 * @param {string} temp Temp directory (absolute)
 * @param {string} build Build directory (absolute)
 * @param {string} file Path to Typst source file
 * @param {string | undefined} deckName Deck name
 */
async function processDeck(src, temp, build, file, deckName = "deck.csv") {
  const start = dayjs();
  const { cards, assets } = await processTypstSourceFile(file);
  await processAssets(src, temp, assets);
  await generateTypstFiles(src, temp, cards);
  await mkdirp(build);
  await renderCards(src, temp, build, cards);

  // Flush all processes out...
  await new Promise((resolve) => setTimeout(resolve, 250));
  logger.info(`Successfully rendered ${cards.length} flashcards!`);

  await writeDeckCsv(build, cards, deckName);
  logger.info(`Deck CSV successfully generated!`);
  logger.debug(`Finished in ${dayjs().diff(start, "s")} seconds.`);
  logger.flush();
}

await configureDirectory("deck");
await configureDirectory("temp");
await configureDirectory("temp/7ccsmsen");
await processDeck(
  "/home/izzy/projects/typst-notes/y4s1",
  resolve("temp"),
  resolve("deck"),
  "/home/izzy/projects/typst-notes/y4s1/7ccsmsen.typ",
  "7ccsmsen.csv"
);
await configureDirectory("temp/7ccsmeai");
await processDeck(
  "/home/izzy/projects/typst-notes/y4s1",
  resolve("temp"),
  resolve("deck"),
  "/home/izzy/projects/typst-notes/y4s1/7ccsmeai.typ",
  "7ccsmeai.csv"
);
await configureDirectory("temp/7ccsmeai-practice");
await processDeck(
  "/home/izzy/projects/typst-notes/y4s1",
  resolve("temp"),
  resolve("deck"),
  "/home/izzy/projects/typst-notes/y4s1/7ccsmeai-practice.typ",
  "7ccsmeai-practice.csv"
);
await configureDirectory("temp/7ccsmams");
await processDeck(
  "/home/izzy/projects/typst-notes/y4s1",
  resolve("temp"),
  resolve("deck"),
  "/home/izzy/projects/typst-notes/y4s1/7ccsmams.typ",
  "7ccsmams.csv"
);
await cleanUp();
